"""
API methods
---
- i/o
    - load
    - save

- subsetting
    - subset_by_dir
    - subset_by_image_ids

- getters
    - get_filenames_from_dir
    - get_ids_from_imgs
    - get_img_by_filename
    - get_img_by_image_id
    - get_ann_by_image_id

- concatenate
- reindex_coco
"""

# native imports
import json
import os
import numpy as np
import pyniche


class COCO_API:
    def __init__(self, data=None, path_json=None):
        # members
        self.data = None  # dict of COCO keys
        self.path_json = path_json

        # init
        if data is not None:
            # 1. user provided dict
            self.data = data
        else:
            # 2. user provided path to json
            self.load()

    def load(self):
        with open(self.path_json, "r") as f:
            self.data = json.load(f)

    def save(self, path):
        with open(path, "w") as f:
            json.dump(self.data, f)

    def new_instance(self, imgs_coco, ann_coco):
        # re-index images and annotations
        imgs_coco, ann_coco = reindex_coco(imgs_coco, ann_coco)

        # new instance
        new_json = dict(
            {
                "info": self.info(),
                "licenses": self.licenses(),
                "categories": self.categories(),
                "images": imgs_coco,
                "annotations": ann_coco,
            }
        )
        return COCO_API(data=new_json)

    # subsetting
    def subset_by_dir(self, dir_img):
        # get images from dir
        filenames = self.get_filenames_from_dir(dir_img)
        imgs_coco = self.get_img_by_filename(filenames)

        # get ann by ids of images
        ids = self.get_ids_from_imgs(imgs_coco)
        ann_coco = self.get_ann_by_image_id(ids)

        # return
        return self.new_instance(imgs_coco, ann_coco)

    def subset_by_image_ids(self, ids):
        if isinstance(ids, int):
            # if int, it's a single id
            ids = [ids]
        elif isinstance(ids, list):
            # if list, it's a list of ids
            ids = ids
        elif isinstance(ids, tuple):
            # if tuple, it's a range
            ids = list(np.arange(ids[0], ids[1] + 1))

        # filter images
        imgs_coco = self.get_img_by_image_id(ids)
        ann_coco = self.get_ann_by_image_id(ids)

        # return
        return self.new_instance(imgs_coco, ann_coco)

    # getters
    def get_filenames_from_dir(self, dir_img):
        filenames = [f for f in os.listdir(dir_img) if ".jpg" in f or ".png" in f]
        return filenames

    def get_ids_from_imgs(self, imgs_coco):
        ids = [i["id"] for i in imgs_coco]
        return ids

    def get_img_by_filename(self, filenames):
        imgs_coco = [i for i in self.images() if i["file_name"] in filenames]
        return imgs_coco

    def get_img_by_image_id(self, ids):
        imgs_coco = [i for i in self.images() if i["id"] in ids]
        return imgs_coco

    def get_ann_by_image_id(self, ids):
        ann_coco = [a for a in self.annotations() if a["image_id"] in ids]
        return ann_coco

    # concat
    def concatenate(self, coco):
        """
        both self and coco are COCO_API instances
        """
        # check the self data and max id
        imgs_self = self.images()
        ann_self = self.annotations()
        max_img_id_self = max([i["id"] for i in imgs_self])
        max_ann_id_self = max([a["id"] for a in ann_self])
        # check the coco data and reinex
        imgs_coco = coco.images()
        ann_coco = coco.annotations()
        imgs_coco, ann_coco = reindex_coco(
            imgs_coco,
            ann_coco,
            base_img=max_img_id_self + 1,
            base_ann=max_ann_id_self + 1,
        )

        # concatenate
        imgs_new = imgs_self + imgs_coco
        ann_new = ann_self + ann_coco
        return self.new_instance(imgs_new, ann_new)

    # COCO keys
    def info(self):
        return self.data["info"]

    def licenses(self):
        return self.data["licenses"]

    def categories(self):
        return self.data["categories"]

    def images(self):
        return self.data["images"]

    def annotations(self):
        return self.data["annotations"]


def reindex_coco(images, annotations, base_img=0, base_ann=0):
    """
    re-index images and annotations, starting from base
    """
    # re-index images
    for i, img in enumerate(images):
        img["id"] = base_img + i
        for ann in annotations:
            if ann["image_id"] == img["id"]:
                ann["image_id"] = img["id"]
    # re-index annotations
    for i, ann in enumerate(annotations):
        ann["id"] = base_ann + i
    return images, annotations


# # IMPLEMENTATION -----
# PATH = os.path.join("/Users", "niche", "_03_Papers", "2024", "cowsformer", "data")


# def get_json(setname):
#     """
#     arguments
#     ---
#     setname: side, top, holstein

#     prerequisite
#     ---
#     PATH is the root path of the data
#     """
#     return os.path.join(PATH, "coco_{}.json".format(setname))


# def get_dir(setname, task):
#     """
#     arguments
#     ---
#     setname: 1a_angle_t2s, 1b_angle_s2t, 2_light, 3_breed, 4_all
#     task: train, test

#     prerequisite
#     ---
#     PATH is the root path of the data
#     """
#     return os.path.join(PATH, setname, task)


# from pyniche.data.COCO.API import COCO_API

# coco_side = COCO_API(PATH_JSON["side"])
# coco_top = COCO_API(PATH_JSON["top"])
# coco_holstein = COCO_API(PATH_JSON["holstein"])

# dir_tgt = get_dir("1a_angle_t2s", "test")
# imgs_tgt = [f for f in os.listdir(dir_tgt) if ".jpg" in f or ".png" in f]
# imgs_tgt

# imgs_coco = [i for i in coco_side.images() if i["file_name"] in imgs_tgt]
# ids_coco = [i["id"] for i in imgs_coco]
# ann_coco = [a for a in coco_side.annotations() if a["image_id"] in ids_coco]

# imgs_coco, ann_coco = reindex_coco(imgs_coco, ann_coco)

# new_json = dict(
#     {
#         "info": coco_side.info(),
#         "licenses": coco_side.licenses(),
#         "categories": coco_side.categories(),
#         "images": imgs_coco,
#         "annotations": ann_coco,
#     }
# )


# len(ann_coco)

# len(imgs_coco)

# len(imgs_tgt)



import lightning as L

from torch.utils.data import random_split, DataLoader
from datasets import concatenate_datasets, load_dataset


class BaseDataModule(L.LightningDataModule):
    """
    parameters
    ---
    dataname and configname: str
        local config file or repository name on the huggingface hub
    batch: int (default: 32, optional)
        batch size
    split_trainval: any (default: None, optional)
        two-element list
        if None, keep the default splits for train/val/test
        if a list of floats, e.g., [0.8, 0.2], then split the train/val with the given ratio
        if a list of ints, e.g., [100, 20], then split the train/val with the given number of samples

    """

    def __init__(
        self,
        dataname: str,
        configname: str = None,
        batch: int = 32,
        split_trainval: any = None,
    ):
        super().__init__()
        # input parameters
        self.dataname = dataname
        self.configname = configname
        self.batch = batch
        self.split_trainval = split_trainval
        # datasets
        self.dataset = {
            "train": None,
            "val": None,
            "test": None,
        }

    def prepare_data(self):
        # download
        pass

    def setup(self, stage: str):
        # Assign train/val datasets for use in dataloaders
        if stage == "fit":
            # load and re-distribute train/val
            data_train = load_dataset(self.dataname, self.configname, split="train")
            data_val = load_dataset(self.dataname, self.configname, split="validation")
            data_train, data_val = self.re_distribute(
                data_train,
                data_val,
                self.split_trainval,
            )
            # assign and set transforms
            self.dataset["train"] = data_train
            self.dataset["val"] = data_val
        if stage == "test":
            data_test = load_dataset(self.dataname, self.configname, split="test")
            self.dataset["test"] = data_test

    # loaders
    def train_dataloader(self):
        return DataLoader(
            self.dataset["train"].with_transform(self._transform_train),
            batch_size=self.batch,
            shuffle=True,
            pin_memory=True,
            collate_fn=self._collate_fn_train,
        )

    def val_dataloader(self):
        return DataLoader(
            self.dataset["val"].with_transform(self._transform_val),
            batch_size=self.batch,
            shuffle=False,
            pin_memory=True,
            collate_fn=self._collate_fn_val,
        )

    def test_dataloader(self):
        return DataLoader(
            self.dataset["test"].with_transform(self._transform_test),
            batch_size=self.batch,
            shuffle=False,
            pin_memory=True,
            collate_fn=self._collate_fn_test,
        )

    # additional methods
    def re_distribute(self, data_train, data_val, split_trainval):
        """
        split_trainval: list, [n_train, n_val]
        """
        if split_trainval is not None:
            # concatenate to get the full dataset
            data_full = concatenate_datasets([data_train, data_val])
            n = len(data_full)

            # check int (count) or float (ratio)
            if isinstance(split_trainval[0], float):
                n_train = int(split_trainval[0] * n)
                n_val = n - n_train
            else:
                n_train, n_val = split_trainval
            # if n_val is not proivded
            if n_val == -1:
                n_val = n - n_train

            # split
            if n_train + n_val == n:
                data_train, data_val = random_split(data_full, split_trainval)
            elif n_train + n_val < n:
                data_train, data_val = random_split(data_full, [n - n_val, n_val])
                data_train, _ = random_split(data_train, [n_train, n - n_train - n_val])
            elif n_train + n_val > n:
                print("The sum of train/val samples is larger than the total samples.")
                raise ValueError

        # return
        return data_train, data_val

    # need to be implemented
    def _collate_fn(self, batch):
        raise NotImplementedError

    def _collate_fn_train(self, batch):
        batch = self._collate_fn(batch)
        return batch

    def _collate_fn_val(self, batch):
        batch = self._collate_fn(batch)
        return batch

    def _collate_fn_test(self, batch):
        batch = self._collate_fn(batch)
        return batch

    def _transform_train(examples):
        return examples

    def _transform_val(examples):
        return examples

    def _transform_test(examples):
        return examples



import matplotlib.pyplot as plt
import matplotlib.patches as patches
import random
import numpy as np
import colorsys


def show_dataset(dataset, idx=None):
    # COCO API
    coco = dataset.coco

    # get image id from idx
    img_ids = list(coco.imgs.keys())
    if idx is None:
        idx = random.randint(0, len(img_ids))
    img_id = img_ids[idx]
    image = dataset._load_image(img_id)

    # Get all distinct categories and create a color for each of them
    categories = coco.loadCats(coco.getCatIds())
    num_classes = len(categories)
    hsv_tuples = [(x / num_classes, 1.0, 1.0) for x in range(num_classes)]
    colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
    colors = [(int(r * 255), int(g * 255), int(b * 255)) for r, g, b in colors]
    random.shuffle(colors)

    # Display the image
    plt.figure(figsize=(10, 10))
    plt.imshow(image)

    # Add the bounding boxes
    ax = plt.gca()
    annotation_ids = coco.getAnnIds(imgIds=img_id, iscrowd=None)
    annotations = coco.loadAnns(annotation_ids)
    annotation = annotations[0]
    annotation["bbox"]
    annotation["segmentation"][0]

    for annotation in annotations:
        # Draw bounding box
        bbox = annotation["bbox"]
        box = patches.Rectangle(
            (bbox[0], bbox[1]),
            bbox[2],
            bbox[3],
            linewidth=2,
            edgecolor="r",
            facecolor="none",
        )
        ax.add_patch(box)

        # Draw segmentation
        if "segmentation" in annotation:
            category_id = annotation["category_id"]
            color = (
                colors[categories.index(category_id)]
                if category_id in categories
                else "r"
            )
            for seg in annotation["segmentation"]:
                poly = np.array(seg).reshape((int(len(seg) / 2), 2))
                polygon = patches.Polygon(
                    poly, linewidth=1, edgecolor=color, facecolor=color, alpha=0.5
                )
                ax.add_patch(polygon)

        # Draw label
        label = coco.loadCats(annotation["category_id"])[0]["name"]
        plt.text(
            bbox[0],
            bbox[1] - 10,
            label,
            color="white",
            fontsize=12,
            bbox={"facecolor": "red", "alpha": 0.5},
        )
